'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

// attempt at making a standalone testable data structure

// we could make razordata and bladedata inherit from a common class
// but honestly didnt want to prematurely optimize

var RazorData = exports.RazorData = function () {
  function RazorData(_ref) {
    var _ref$args = _ref.args,
        args = _ref$args === undefined ? null : _ref$args,
        _ref$name = _ref.name,
        name = _ref$name === undefined ? null : _ref$name,
        _ref$type = _ref.type,
        type = _ref$type === undefined ? null : _ref$type,
        _ref$fragmentType = _ref.fragmentType,
        fragmentType = _ref$fragmentType === undefined ? null : _ref$fragmentType;

    _classCallCheck(this, RazorData);

    if (!type) throw new Error('type must be either fragment or query');
    if (type === 'fragment' && !fragmentType) throw new Error('fragments must come with a fragmentType');
    if (type === 'fragment' && !name) throw new Error('fragments must come with a name');
    this._children = []; // all the blades
    this._args = args; // a string for now
    this._name = name; // truly optional
    this._type = type; // either 'fragment' or 'query'
    this._fragmentType = fragmentType; // if fragment
  }

  RazorData.prototype.isFragment = function isFragment() {
    return this._type === 'fragment';
  };

  RazorData.prototype.getFragmentData = function getFragmentData() {
    return {
      name: this._name,
      fragmentType: this._fragmentType
    };
  };

  RazorData.prototype.get = function get(id) {
    for (var i = 0; i < this._children.length; i++) {
      var name = this._children[i]._name === id;
      var alias = this._children[i]._alias === id;
      if (name || alias) return this._children[i];
    }
    return null;
  };

  RazorData.prototype.add = function add(val) {
    var child = this.get(val.name);
    if (!(child && child._alias == val.alias)) {
      child = new BladeData(val);
      this._children.push(child);
    }
    return child; // return child so that more operations can be made
  };

  RazorData.prototype.print = function print() {
    var fields = this._children;
    if (!fields.length) return console.log('babel-blade Warning: razor with no children, doublecheck') || null; // really shouldnt happen, should we throw an error?
    var maybeArgs = this._args && this._args[0];
    var TemplateLiteral = appendLiterals().addStr(this._type === 'query' ? `\nquery ${this._name || ''}` : `\nfragment ${this._name} on ${this._fragmentType}`).addStr(maybeArgs ? '(' : '').addLit(maybeArgs).addStr(maybeArgs ? ')' : '').addStr('{\n');

    var fragments = []; // will be mutated to add all the fragments included
    var accumulators = Object.keys(fields).map(function (key) {
      return fields[key].print('  ', fragments);
    });
    accumulators.forEach(TemplateLiteral.append);
    TemplateLiteral.addStr('}'); // cap off the string
    // todo: append to accumulator
    fragments.forEach(function (frag) {
      return TemplateLiteral.append(frag.print());
    });
    var g = TemplateLiteral.get();
    // console.log({ g });
    return zipAccumulators(g);
  };

  return RazorData;
}();

var BladeData = exports.BladeData = function () {
  function BladeData(_ref2) {
    var _ref2$name = _ref2.name,
        name = _ref2$name === undefined ? null : _ref2$name,
        _ref2$args = _ref2.args,
        args = _ref2$args === undefined ? null : _ref2$args,
        _ref2$fragments = _ref2.fragments,
        fragments = _ref2$fragments === undefined ? [] : _ref2$fragments,
        _ref2$directive = _ref2.directive,
        directive = _ref2$directive === undefined ? null : _ref2$directive;

    _classCallCheck(this, BladeData);

    if (!name) throw new Error('new Blade must have name');
    if (!Array.isArray(fragments)) throw new Error('fragments must be array');
    this._children = []; // store of child blades
    this._name = name; // a string for now
    this._args = args; // a string for now
    this._alias = args && `${this._name}_${hashCode(JSON.stringify(args))}`;
    this._fragments = fragments;
    this._directive = directive;

    // binding cos im too lazy to set up public class fields
    // this.get = this.get.bind(this);
    // this.add = this.add.bind(this);
    // this.print = this.print.bind(this);
  }

  BladeData.prototype.get = function get(id) {
    for (var i = 0; i < this._children.length; i++) {
      if (this._children[i]._name === id) return this._children[i];
    }
    return null;
  };

  BladeData.prototype.add = function add(val) {
    var child = this.get(val.name);
    if (!(child && child._alias == val.alias)) {
      child = new BladeData(val);
      this._children.push(child);
    }
    return child;
  };
  // TODO: potential problem here if blade has args/alias but no children


  BladeData.prototype.print = function print(indent, fragments) {
    var maybeArgs = this._args && this._args[0];
    var alias = this._alias;
    var printName = alias ? `${alias}: ${this._name}` : this._name;
    var printDirectives = this._directive ? ` ${this._directive}` : '';
    if (this._fragments.length) this._fragments.map(function (frag) {
      return fragments.push(frag);
    }); // mutates fragments!
    // let GraphQLString = `${indent}${printName}${maybeArgs}${printDirectives}`;
    var TemplateLiteral = appendLiterals().addStr(`${indent}${printName}`).addStr(maybeArgs ? '(' : '').addLit(maybeArgs).addStr(maybeArgs ? ')' : '').addStr(printDirectives);
    var fields = this._children;
    if (fields.length || this._fragments.length) {
      TemplateLiteral.addStr(' {\n');
      var accumulators = Object.keys(fields).map(function (key) {
        return fields[key].print(indent + '  ', fragments);
      });
      accumulators.forEach(TemplateLiteral.append);
      this._fragments.forEach(function (frag) {
        TemplateLiteral.addStr(`${indent}  ...${frag.getFragmentData().name}\n`);
      });
      TemplateLiteral.addStr(`${indent}}\n`); // cap off the query
    } else {
      TemplateLiteral.addStr('\n');
    }
    return TemplateLiteral.get();
  };

  return BladeData;
}();

function generateFields(obj, references) {
  var indent = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : '  ';

  return indent + Object.keys(obj).map(function (key) {
    var name = key;
    if (obj[key] && obj[key] && Object.keys(obj[key]).length && obj[key].isReference()) {
      name = `...${key[0].toLowerCase() + key.slice(1) + 'Fragment'}`;
      references.add(key);
    }
    return `${name}` + (obj[key] && obj[key] && Object.keys(obj[key]).length ? ` {\n` + generateFields(obj[key], references, indent + '  ') + `\n${indent}}` : '');
  }).join(`\n${indent}`);
}

// function generateTemplate(s, references) {
// 	return t.templateLiteral(
// 		//[t.templateElement({raw: '\n'}), t.templateElement({raw: s})],
// 		[t.templateElement({ raw: s })],
// 		[...references].map(ref => t.memberExpression(t.identifier(ref), t.identifier('fragment')))
// 	);
// }

// https://stackoverflow.com/a/8831937/1106414
function hashCode(str) {
  var hash = 0;
  if (str.length == 0) {
    return hash;
  }
  for (var i = 0; i < str.length; i++) {
    var char = str.charCodeAt(i);
    hash = (hash << 5) - hash + char;
    hash = hash & hash; // Convert to 32bit integer
  }
  return hash.toString(16).slice(-4); // last4hex
}

function appendLiterals() {
  var stringAccumulator = [];
  var litAccumulator = [];
  var me = {
    addStr() {
      var str = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

      stringAccumulator.push(str);
      litAccumulator.push(null);
      return me;
    },
    addLit() {
      var lit = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;

      stringAccumulator.push(null);
      litAccumulator.push(lit);
      return me;
    },
    add() {
      var str = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : null;
      var lit = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : null;

      stringAccumulator.push(str);
      litAccumulator.push(lit);
      return me;
    },
    append(newMe) {
      newMe.stringAccumulator.forEach(function (str) {
        return stringAccumulator.push(str);
      });
      newMe.litAccumulator.forEach(function (lit) {
        return litAccumulator.push(lit);
      });
      return me;
    },
    get() {
      return { stringAccumulator, litAccumulator };
    }
  };
  return me;
}

function zipAccumulators(_ref3) {
  var stringAccumulator = _ref3.stringAccumulator,
      litAccumulator = _ref3.litAccumulator;

  // cannot have any spare
  var str = '',
      newStrAcc = [],
      newLitAcc = [];
  for (var i = 0; i < stringAccumulator.length; i++) {
    if (litAccumulator[i]) {
      var maybeSimpleString = maybeGetSimpleString(litAccumulator[i]);
      if (maybeSimpleString) {
        // its just a simplestring!
        str += maybeSimpleString;
      } else {
        newLitAcc.push(litAccumulator[i]);
        newStrAcc.push(str + (stringAccumulator[i] || ''));
        str = '';
      }
    } else {
      // there is an empty lit, store in state
      str += stringAccumulator[i] || '';
    }
  }
  // flush store
  if (str !== '') newStrAcc.push(str);
  return { stringAccumulator: newStrAcc, litAccumulator: newLitAcc };
}

function maybeGetSimpleString(Literal) {
  if (Literal.type === 'StringLiteral') return Literal.value;
  if (Literal.type === 'TemplateLiteral' && !Literal.expressions.length && Literal.quasis.length === 1) return Literal.quasis[0].value.raw;
  // else
  return null;
}